/*
 * Copyright (C) 2023 Xeme Developers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <https://www.gnu.org/licenses/>.
 */

#include <birb.h>

#include "xemecore.h"
#include "xemeoutputstream.h"
#include "xemestring.h"

struct _XemeOutputStream {
	XemeStream parent;

	gboolean running;

	GOutputStream *output;
};

G_DEFINE_FINAL_TYPE(XemeOutputStream, xeme_output_stream, XEME_TYPE_STREAM)

/******************************************************************************
 * Helpers
 *****************************************************************************/
static GBytes *
xeme_output_stream_serialize(XemeOutputStream *output_stream) {
	XemeStream *stream = XEME_STREAM(output_stream);
	GString *str = NULL;
	const char *value = NULL;

	str = g_string_new(XEME_STREAM_PI);
	g_string_append(str, "<stream:stream");

	/* to and from are validated to not be empty in
	 * xeme_output_stream_start. */
	g_string_append_printf(str, " from=\"%s\"", xeme_stream_get_from(stream));
	g_string_append_printf(str, " to=\"%s\"", xeme_stream_get_to(stream));

	value = xeme_stream_get_id(stream);
	if(!xeme_str_is_empty(value)) {
		g_string_append_printf(str, " id=\"%s\"", value);
	}

	value = xeme_stream_get_version(stream);
	if(!xeme_str_is_empty(value)) {
		g_string_append_printf(str, " version=\"%s\"", value);
	}

	value = xeme_stream_get_language(stream);
	if(!xeme_str_is_empty(value)) {
		g_string_append_printf(str, " xml:lang=\"%s\"", value);
	}

	g_string_append_printf(str, " xmlns=\"%s\"", XEME_STREAM_XMLNS);
	g_string_append_printf(str, " xmlns:stream=\"%s\"",
	                       XEME_STREAM_XMLNS_STREAM);
	g_string_append(str, ">");

	/* Create a GBytes with the value from the GString and free the GString
	 * container.
	 */
	return g_string_free_to_bytes(str);
}

/******************************************************************************
 * Writing
 *****************************************************************************/
static void
xeme_output_stream_write_cb(G_GNUC_UNUSED GObject *source, GAsyncResult *res,
                            gpointer data)
{
	XemeOutputStream *stream = data;
	BirbQueuedOutputStream *queued = NULL;
	GError *error = NULL;

	queued = BIRB_QUEUED_OUTPUT_STREAM(stream->output);
	birb_queued_output_stream_push_bytes_finish(queued, res, &error);

	if(error != NULL) {
		/* TODO: We need to propagate this somehow, maybe a signal? */
		g_warning("failed to write data: %s",
		          error->message != NULL ? error->message : "unknown error");
	}
}

static void
xeme_output_stream_write(XemeOutputStream *stream, GBytes *data) {
	BirbQueuedOutputStream *queued = NULL;
	GCancellable *cancellable = NULL;

	g_return_if_fail(XEME_IS_OUTPUT_STREAM(stream));
	g_return_if_fail(data != NULL);

	cancellable = xeme_stream_get_cancellable(XEME_STREAM(stream));
	queued = BIRB_QUEUED_OUTPUT_STREAM(stream->output);

	birb_queued_output_stream_push_bytes_async(queued,
	                                           data,
	                                           G_PRIORITY_DEFAULT,
	                                           cancellable,
	                                           xeme_output_stream_write_cb,
	                                           stream);
}

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
static void
xeme_output_stream_finalize(GObject *obj) {
	XemeOutputStream *stream = XEME_OUTPUT_STREAM(obj);

	g_clear_object(&stream->output);

	G_OBJECT_CLASS(xeme_output_stream_parent_class)->finalize(obj);
}

static void
xeme_output_stream_init(XemeOutputStream *stream) {
	stream->running = FALSE;
}

static void
xeme_output_stream_class_init(XemeOutputStreamClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);

	obj_class->finalize = xeme_output_stream_finalize;
}

/******************************************************************************
 * Public API
 *****************************************************************************/
XemeOutputStream *
xeme_output_stream_new(const char *to, const char *from,
                       GCancellable *cancellable)
{
	return g_object_new(
		XEME_TYPE_OUTPUT_STREAM,
		"cancellable", cancellable,
		"from", from,
		"to", to,
		NULL);
}

gboolean
xeme_output_stream_start(XemeOutputStream *stream, GOutputStream *output,
                         GError **error)
{
	GBytes *data = NULL;
	GOutputStream *wrapper = NULL;

	g_return_val_if_fail(XEME_IS_OUTPUT_STREAM(stream), FALSE);
	g_return_val_if_fail(G_IS_OUTPUT_STREAM(output), FALSE);
	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);

	if(stream->running) {
		g_set_error_literal(error, XEME_DOMAIN, 0,
		                    "output stream has already been started");

		return FALSE;
	}

	wrapper = birb_queued_output_stream_new(output);
	if(!g_set_object(&stream->output, wrapper)) {
		g_set_error_literal(error, XEME_DOMAIN, 0,
		                    "failed to set the output stream");

		g_clear_object(&wrapper);

		return FALSE;
	}

	/* The wrapper is referenced by stream->output so we need to get rid of our
	 * local reference which we don't need anymore.
	 */
	g_clear_object(&wrapper);

	data = xeme_output_stream_serialize(stream);
	xeme_output_stream_write(stream, data);
	g_clear_pointer(&data, g_bytes_unref);

	stream->running = TRUE;

	return TRUE;
}

gboolean
xeme_output_stream_restart(XemeOutputStream *stream, GError **error) {
	GBytes *data = NULL;

	g_return_val_if_fail(XEME_IS_OUTPUT_STREAM(stream), FALSE);
	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);

	if(!stream->running) {
		g_set_error_literal(error, XEME_DOMAIN, 0,
		                    "output stream has not been started");

		return FALSE;
	}

	data = xeme_output_stream_serialize(stream);
	xeme_output_stream_write(stream, data);
	g_clear_pointer(&data, g_bytes_unref);

	return TRUE;
}

gboolean
xeme_output_stream_stop(XemeOutputStream *stream, GError **error) {
	GCancellable *cancellable = NULL;
	GError *local_error = NULL;
	gboolean ret = TRUE;

	g_return_val_if_fail(XEME_IS_OUTPUT_STREAM(stream), FALSE);

	cancellable = xeme_stream_get_cancellable(XEME_STREAM(stream));

	if(!g_output_stream_close(stream->output, cancellable, &local_error)) {
		g_propagate_error(error, local_error);

		ret = FALSE;
	}

	stream->output = NULL;

	return ret;
}
